{-|
Module      : Majority Vote
Description : Majority vote
Copyright   : (c) CNRS, 2018-Present
License     : AGPL + CECILL v3
Maintainer  : alexandre+dev@delanoe.org
Stability   : experimental
Portability : POSIX

-}

{-# LANGUAGE OverloadedStrings #-}

module Lib where

import Data.Either
import Safe (headMay)
import Data.Maybe (fromJust)
import Data.List
import Data.Text (Text)
import Data.Map (Map)
import qualified Data.Map as DM
import Prelude

------------------------------------------------------
-- | Majority Judgment Elections suppose a list of choices and a Ballot for each choice.
-- TODO: List should be a Set
type MajorityJudgment = [Ballot]

-- | Definition of a ballot
data Ballot = Ballot { label :: Text
                     , votes :: [Vote]
}

instance Show Ballot where
  show (Ballot l vs) = show (l, toMap vs)

data Vote = Poor | Fair | Good | Excellent
  deriving (Eq, Ord, Enum, Bounded, Show)

------------------------------------------------------------------------
-- | Example application
-- https://en.wikipedia.org/wiki/Majority_judgment
ballots :: MajorityJudgment
ballots = [ Ballot "Memphis"     (ratedAs [(42, Excellent),(26, Poor)     , (15, Poor)     , (17, Poor)])
          , Ballot "Nashville"   (ratedAs [(42, Fair     ),(26, Excellent), (15, Fair)     , (17, Fair)])
          , Ballot "Chattanooga" (ratedAs [(42, Poor     ),(26, Fair)     , (15, Excellent), (17, Good)])
          , Ballot "Knoxville"   (ratedAs [(42, Poor     ),(26, Fair)     , (15, Good)     , (17, Excellent)])
          ]

-- | Ballots2 test
-- If voters were more strategic, those from Knoxville and Chattanooga
-- might rate Nashville as "Poor" and Chattanooga as "Excellent", in
-- an attempt to make their preferred candidate Chattanooga win. Also,
-- Nashville voters might rate Knoxville as "poor" to distinguish it from
-- Chattanooga. In spite of these attempts at strategy, the winner would
-- still be Nashville
ballots2 :: MajorityJudgment
ballots2 = [ Ballot "Memphis"     (ratedAs [(42, Excellent),(26, Poor)     , (15, Poor)     , (17, Poor)])
           , Ballot "Nashville"   (ratedAs [(42, Fair     ),(26, Excellent), (15, Poor)     , (17, Poor)])
           , Ballot "Chattanooga" (ratedAs [(42, Poor     ),(26, Poor)     , (15, Excellent), (17, Good)])
           , Ballot "Knoxville"   (ratedAs [(42, Poor     ),(26, Poor)     , (15, Excellent), (17, Excellent)])
           ]

ratedAs = concat . map (\(n,v) -> (take n . repeat) v)

-- | Unfair election
ballots3 :: MajorityJudgment
ballots3 = [ Ballot "Dictatorship" (ratedAs [(1, Excellent)])]

ballots4 :: MajorityJudgment
ballots4 = [ Ballot "Fascism"   (ratedAs [(90, Excellent)])
           , Ballot "Democracy" (ratedAs [(10, Excellent), (90, Poor)])
           ]

------------------------------------------------------------------------

-- | Select the Winner
-- TODO: sortBy and intersect according to Result
-- sort then take While equality
winnerIs :: MajorityJudgment -> Either Unfair (Maybe Ballot)
winnerIs mj = case isFair mj of
                Just unfair -> Left  unfair
                Nothing     -> Right $ (headMay . reverse . sort) mj

data Unfair = NoChoice | NoRepresentativity
  deriving (Show)

-- | Asserts is the Election with Majority Judgment is Fair
-- e.g. methodology of vote is respected to collect the ballots
-- Number of vote per Ballot must be equal
-- add an axiom: must not be Poor at least by one voter ?
-- TODO: be more explicit why election is not fair
isFair :: MajorityJudgment -> Maybe Unfair
isFair mj = case length vs < 2 of
  True -> Just NoChoice
  False -> case all (== head vs) vs of
    False -> Just NoRepresentativity
    True  -> Nothing
  where
    vs = map (length . votes) mj



instance Eq Ballot where
  (==) (Ballot _ v1) (Ballot _ v2) = (==) (score v1) (score v2)

instance Ord Ballot where
  (compare) (Ballot _ v1) (Ballot _ v2) = case (==) (score v1) (score v2) of
      False -> compare (score v1) (score v2)
      True  -> compare (score v1') (score v2')
            where
              (v1', v2') = xor (v1, v2)

score :: [Vote] -> Maybe Rational
score = median . map voteToRational

------------------------------------------------------------------------

xor :: Ord a => ([a],[a]) -> ([a],[a])
xor (m1,m2) = (fromMap m1', fromMap m2')
  where
    (m1', m2') = xorMap (toMap m1, toMap m2)

xorMap :: Ord a => (Map a Int, Map a Int) -> (Map a Int, Map a Int)
xorMap (m1,m2) = (m1',m2')
  where
    m1' = DM.unionWith outer m1 $ DM.intersectionWith (-) m1 m2
    m2' = DM.unionWith outer m2 $ DM.intersectionWith (-) m2 m1

    outer a b = if b > 0 then b else 0

toMap :: Ord a => [a] -> Map a Int
toMap = foldl' (\m a -> DM.insertWith (+) a 1 m) DM.empty

fromMap :: Map a Int -> [a]
fromMap = concat . map (\(k,n) -> (take n . repeat) k) . DM.toList

------------------------------------------------------------------------
-- | TODO Make instance ScoreVote
voteToRational :: (Ord k, Bounded k, Enum k) => k -> Rational
voteToRational v = fromJust $ DM.lookup v
                            $ DM.fromList
                            $ zip vs  [0 .. (toRational . length) vs]
  where
    vs = [minBound..maxBound]

medianFromSorted :: Fractional a => [a] -> Maybe a
medianFromSorted [] = Nothing
medianFromSorted [a] = Just a
medianFromSorted [a,b] = Just ((a + b) / 2)
medianFromSorted (a:xs) = medianFromSorted (init xs) -- init is not efficient

median :: Ord a => Fractional a => [a] -> Maybe a
median = medianFromSorted . sort

